package core.api.controller;

import core.api.models.auth.LoginRequest;
import core.api.models.auth.LoginResponse;
import core.api.models.auth.SignupRequest;
import core.api.models.auth.SimpleStatusResponse;
import core.entities.RoleEntity;
import core.entities.User;
import core.entities.UserRole;
import core.repositories.RoleRepository;
import core.repositories.UserRepository;
import core.security.jwt.JWTUtils;
import core.security.service.UserDetailsImpl;
import core.services.PropertyService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.authentication.AuthenticationManager;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.web.bind.annotation.*;

import javax.management.relation.Role;
import javax.validation.Valid;
import java.util.*;
import java.util.stream.Collectors;

@CrossOrigin(origins = "*", maxAge = 3600)
@RestController
@RequestMapping("/api/auth")
public class AuthController {

    @Autowired
    AuthenticationManager authenticationManager;
    @Autowired
    UserRepository userRepository;
    @Autowired
    RoleRepository roleRepository;
    @Autowired
    private PropertyService propertyService;

    PasswordEncoder encoder = new BCryptPasswordEncoder();
    @Autowired
    JWTUtils jwtUtils;

    @PostMapping("/signin")
    public ResponseEntity<?> authenticateUser(@Valid @RequestBody LoginRequest loginRequest) {
        Authentication authentication = authenticationManager.authenticate(
                new UsernamePasswordAuthenticationToken(loginRequest.getUsername(), loginRequest.getPassword()));
        SecurityContextHolder.getContext().setAuthentication(authentication);


        UserDetailsImpl userDetails = (UserDetailsImpl) authentication.getPrincipal();
        List<String> roles = userDetails.getAuthorities().stream()
                .map(GrantedAuthority::getAuthority).toList();
        boolean admin = roles.contains("ROLE_ADMIN");
        String jwt = jwtUtils.generateJWTToken(authentication, admin);
        return ResponseEntity.ok(new LoginResponse("successful", jwt));
    }

    @PutMapping("/signup")
    public ResponseEntity<?> registerUser(@Valid @RequestBody SignupRequest signUpRequest) {
        if(!propertyService.isPropertyEnabled(PropertyService.PropertyName.REGISTRATION_PROPERTY_NAME)) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if (userRepository.existsByUsername(signUpRequest.getUsername())) {
            return ResponseEntity
                    .status(HttpStatus.CONFLICT)
                    .body(new SimpleStatusResponse("failed"));
        }
        if (userRepository.existsByEmail(signUpRequest.getEmail())) {
            return ResponseEntity
                    .status(HttpStatus.CONFLICT)
                    .body(new SimpleStatusResponse("failed"));
        }
        // Create new user's account
        User user = new User(signUpRequest.getUsername(),
                signUpRequest.getEmail(),
                encoder.encode(signUpRequest.getPassword()));

        long amountOfUsers = userRepository.countUsers();

        HashSet<RoleEntity> roles = new HashSet<>();
        if(amountOfUsers == 0) {
            Optional<RoleEntity> adminRoleEntity= roleRepository.findByName(UserRole.ROLE_ADMIN);
            adminRoleEntity.ifPresent(roles::add);
        }
        Optional<RoleEntity> userRoleReq = roleRepository.findByName(UserRole.ROLE_USER);
        userRoleReq.ifPresent(roles::add);

        userRoleReq.ifPresent(userRole -> user.setRoles(roles));

        userRepository.save(user);
        return ResponseEntity.ok(new SimpleStatusResponse("success"));
    }
}
