package core.api.controller;


import core.api.models.auth.SimpleStatusResponse;
import core.api.models.timemanager.taskSchedule.*;
import core.api.models.timemanager.taskSchedule.scheduleInfos.*;
import core.entities.timemanager.AbstractSchedule;
import core.entities.timemanager.AdvancedTaskSchedule;
import core.entities.timemanager.BasicTaskSchedule;
import core.entities.timemanager.Task;
import core.services.*;
import core.services.ntfy.TaskSchedulingService;
import org.quartz.SchedulerException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.ResponseEntity;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.web.bind.annotation.*;

import javax.validation.Valid;
import java.time.LocalDate;
import java.util.ArrayList;
import java.util.List;


@CrossOrigin(origins = "*", maxAge = 3600)
@RestController
@RequestMapping("/api")
public class ScheduleController {

    @Autowired private TaskScheduleService taskScheduleService;
    @Autowired private TaskService taskService;
    @Autowired private TaskSchedulingService taskSchedulingService;
    @GetMapping("/schedules")
    public ResponseEntity<?> loadAllSchedulesOfUser() {
        String username = SecurityContextHolder.getContext().getAuthentication().getName();
        List<AbstractSchedule> schedules = taskScheduleService.getAllSchedulesOfUser(username);

        return ResponseEntity.ok(schedules.stream().map(AbstractSchedule::toScheduleInfo).toList());
    }

    @GetMapping("/schedules/{taskID}")
    public ResponseEntity<?> loadAllSchedulesOfTask(@PathVariable long taskID) {
        PermissionResult<Task> permissionResult = taskService.getTaskPermissions(taskID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!permissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(permissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        List<AbstractSchedule> taskSchedules = permissionResult.getResult().getBasicTaskSchedules();
        List<ScheduleInfo> scheduleInfos = new ArrayList<>();
        for(AbstractSchedule schedule : taskSchedules) {
            //Filter completed schedules out
            if(!schedule.isCompleted()) {
                scheduleInfos.add(schedule.toScheduleInfo());
            }
        }
        return ResponseEntity.ok(scheduleInfos);
    }

    @PutMapping("/schedules/{taskID}/basic")
    public ResponseEntity<?> createBasicSchedule(@PathVariable long taskID, @RequestBody @Valid BasicScheduleFieldInfo scheduleFieldInfo) {
        return createAbstractSchedule(taskID, scheduleFieldInfo);
    }

    @PutMapping("/schedules/{taskID}/advanced")
    public ResponseEntity<?> createAdsvancedSchedule(@PathVariable long taskID, @RequestBody @Valid AdvancedScheduleFieldInfo scheduleFieldInfo) {
        return createAbstractSchedule(taskID, scheduleFieldInfo);
    }

    private ResponseEntity<?> createAbstractSchedule(long taskID, ScheduleFieldInfo scheduleFieldInfo) {
        PermissionResult<Task> permissionResult = taskService.getTaskPermissions(taskID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!permissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(permissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        if(scheduleFieldInfo instanceof BasicScheduleFieldInfo) {
            ServiceResult<AbstractSchedule> scheduleResult = taskScheduleService.scheduleBasic(permissionResult.getResult(), (BasicScheduleFieldInfo) scheduleFieldInfo);
            if(scheduleResult.getExitCode() == ServiceExitCode.INVALID_OPERATION) {
                return ResponseEntity.status(400).body(new SimpleStatusResponse("failed"));
            }
            return ResponseEntity.ok(scheduleResult.getResult().toScheduleInfo());
        } else if(scheduleFieldInfo instanceof AdvancedScheduleFieldInfo) {
            ServiceResult<AbstractSchedule> scheduleResult = taskScheduleService.scheduleAdvanced(permissionResult.getResult(), (AdvancedScheduleFieldInfo) scheduleFieldInfo);
            if(scheduleResult.getResult() != null) {
                return ResponseEntity.ok(scheduleResult.getResult().toScheduleInfo());
            } else {
                return ResponseEntity.status(400).body(new SimpleStatusResponse("failed"));
            }

        } else {
            return ResponseEntity.status(400).body(new SimpleStatusResponse("failed"));
        }
    }

    @PostMapping("/schedules/{scheduleID}/basic")
    public ResponseEntity<?> editBasicSchedule(@PathVariable long scheduleID, @RequestBody @Valid BasicScheduleFieldInfo scheduleFieldInfo) {
        return editAbstractSchedule(scheduleID, scheduleFieldInfo);
    }

    @PostMapping("/schedules/{scheduleID}/advanced")
    public ResponseEntity<?> editAdvancedSchedule(@PathVariable long scheduleID, @RequestBody @Valid AdvancedScheduleFieldInfo scheduleFieldInfo) {
        return editAbstractSchedule(scheduleID, scheduleFieldInfo);
    }

    private ResponseEntity<?> editAbstractSchedule(long scheduleID, ScheduleFieldInfo scheduleFieldInfo) {
        PermissionResult<AbstractSchedule> permissionResult = taskScheduleService.getSchedulePermissions(scheduleID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!permissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(permissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        if(permissionResult.getResult() instanceof BasicTaskSchedule && scheduleFieldInfo instanceof BasicScheduleFieldInfo) {
            ServiceResult<AbstractSchedule> updatedSchedule = taskScheduleService.editBasicSchedule((BasicTaskSchedule) permissionResult.getResult(), (BasicScheduleFieldInfo) scheduleFieldInfo);
            if(updatedSchedule.getExitCode() == ServiceExitCode.INVALID_OPERATION) {
                return ResponseEntity.status(400).body(new SimpleStatusResponse("failed"));
            } else {
                return ResponseEntity.ok(updatedSchedule.getResult().toScheduleInfo());
            }
        } else if(permissionResult.getResult() instanceof AdvancedTaskSchedule && scheduleFieldInfo instanceof AdvancedScheduleFieldInfo) {
            ServiceResult<AbstractSchedule> updatedSchedule = taskScheduleService.editAdvancedSchedule((AdvancedTaskSchedule) permissionResult.getResult(), (AdvancedScheduleFieldInfo) scheduleFieldInfo);
            if(updatedSchedule.getExitCode() == ServiceExitCode.INVALID_OPERATION) {
                return ResponseEntity.status(400).body(new SimpleStatusResponse("failed"));
            } else {
                return ResponseEntity.ok(updatedSchedule.getResult().toScheduleInfo());
            }
        }

        return ResponseEntity.status(400).body(new SimpleStatusResponse("failed"));
    }

    @DeleteMapping("/schedules/{scheduleID}")
    public ResponseEntity<?> deleteSchedule(@PathVariable long scheduleID) {
        PermissionResult<AbstractSchedule> permissionResult = taskScheduleService.getSchedulePermissions(scheduleID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!permissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(permissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        taskScheduleService.deleteSchedule(permissionResult.getResult());
        return ResponseEntity.ok(new SimpleStatusResponse("success"));
    }

    @GetMapping("/schedules/{date}/{startable}")
    public ResponseEntity<?> loadFilteredSchedulesOfUser(@PathVariable String date, @PathVariable boolean startable) {
        List<AbstractSchedule> abstractSchedules = taskScheduleService.getFilteredScheduledOfUser(LocalDate.parse(date), startable, SecurityContextHolder.getContext().getAuthentication().getName());

        return ResponseEntity.ok(abstractSchedules.stream().map(AbstractSchedule::toScheduleInfo).toList());
    }

    @PostMapping("/schedules/{taskID}/now")
    public ResponseEntity<?> scheduleNow(@PathVariable long taskID) {
        PermissionResult<Task> permissionResult = taskService.getTaskPermissions(taskID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!permissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(permissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        ServiceResult<AbstractSchedule> scheduleResult = taskScheduleService.scheduleNow(permissionResult.getResult());
        if(scheduleResult.getExitCode() == ServiceExitCode.ENTITY_ALREADY_EXIST) {
            return ResponseEntity.status(409).body(new SimpleStatusResponse("failed"));
        } else {
            taskSchedulingService.sendRunningTaskNotification(scheduleResult.getResult());
            return ResponseEntity.ok(scheduleResult.getResult().toScheduleInfo());
        }
    }

    @GetMapping("/schedules/active")
    public ResponseEntity<?> getActiveSchedule() {
        ServiceResult<AbstractSchedule> serviceResult = taskScheduleService.getActiveSchedule(SecurityContextHolder.getContext().getAuthentication().getName());
        if(serviceResult.getExitCode() == ServiceExitCode.OK) {
            return ResponseEntity.ok(serviceResult.getResult().toScheduleInfo());
        } else {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }
    }

    @PostMapping("/schedules/{scheduleID}/activate")
    public ResponseEntity<?> activateSchedule(@PathVariable long scheduleID) {
        PermissionResult<AbstractSchedule> permissionResult = taskScheduleService.getSchedulePermissions(scheduleID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!permissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(permissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        ServiceResult<AbstractSchedule> serviceResult = taskScheduleService.activateSchedule(permissionResult.getResult());
        if(serviceResult.getResult() instanceof AdvancedTaskSchedule) {
            try {
                taskSchedulingService.stopStartReminderNotification((AdvancedTaskSchedule) serviceResult.getResult());

            } catch (SchedulerException e) {
                throw new RuntimeException(e);
            }
        }
        taskSchedulingService.sendRunningTaskNotification(serviceResult.getResult());
        return ResponseEntity.ok(new ScheduleActivateResponse(serviceResult.getResult().getStartTime()));
    }

    @PostMapping("/schedules/{scheduleID}/stop/{finish}")
    public ResponseEntity<?> stopSchedule(@PathVariable long scheduleID, @PathVariable boolean finish) {
        PermissionResult<AbstractSchedule> permissionResult = taskScheduleService.getSchedulePermissions(scheduleID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!permissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(permissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        ServiceResult<AbstractSchedule> serviceResult = taskScheduleService.stopSchedule(permissionResult.getResult(), finish);
        return ResponseEntity.ok(new TaskScheduleStopResponse(serviceResult.getResult().getActiveTime()));
    }

    @PostMapping("/schedules/{taskID}/forgotten")
    public ResponseEntity<?> registerForgottenSchedule(@PathVariable long taskID, @RequestBody @Valid ForgottenScheduleInfo forgottenScheduleInfo) {
        PermissionResult<Task> permissionResult = taskService.getTaskPermissions(taskID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!permissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(permissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        ServiceResult<AbstractSchedule> scheduleResult = taskScheduleService.registerForgottenSchedule(permissionResult.getResult(), forgottenScheduleInfo);
        if(scheduleResult.getExitCode() == ServiceExitCode.INVALID_OPERATION) {
            return ResponseEntity.status(400).body(new SimpleStatusResponse("failed"));
        } else {
            return ResponseEntity.ok(new TaskScheduleStopResponse(scheduleResult.getResult().getActiveTime()));
        }
    }

    @GetMapping("/schedules/missed")
    public ResponseEntity<?> loadMissedSchedules() {
        List<AbstractSchedule> missedSchedules = taskScheduleService.getAllMissedSchedulesOfUser(SecurityContextHolder.getContext().getAuthentication().getName());
        return ResponseEntity.ok(missedSchedules.stream().map(AbstractSchedule::toScheduleInfo).toList());
    }

    @DeleteMapping("/schedules/{scheduleIDs}/all")
    public ResponseEntity<?> deleteSchedules(@PathVariable long[] scheduleIDs) {
        List<PermissionResult<AbstractSchedule>> permissionResults = new ArrayList<>();
        for(long scheduleID: scheduleIDs) {
            PermissionResult<AbstractSchedule> permissionResult = taskScheduleService.getSchedulePermissions(scheduleID, SecurityContextHolder.getContext().getAuthentication().getName());
            if(!permissionResult.isHasPermissions()) {
                return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
            }

            if(permissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
                return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
            }
            permissionResults.add(permissionResult);
        }

        for(PermissionResult<AbstractSchedule> permissionResult : permissionResults) {
            taskScheduleService.deleteSchedule(permissionResult.getResult());
        }
        return ResponseEntity.ok(new SimpleStatusResponse("success"));
    }

    @GetMapping("/schedules/{scheduleID}/details")
    public ResponseEntity<?> loadScheduleDetails(@PathVariable long scheduleID) {
        PermissionResult<AbstractSchedule> permissionResult = taskScheduleService.getSchedulePermissions(scheduleID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!permissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(permissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        return ResponseEntity.ok(permissionResult.getResult().toScheduleInfo());
    }

    @PostMapping("/schedules/{scheduleID}/stopManual")
    public ResponseEntity<?> stopManual(@PathVariable long scheduleID, @Valid @RequestBody ManualScheduleStopInfo manualScheduleStopInfo) {
        PermissionResult<AbstractSchedule> permissionResult = taskScheduleService.getSchedulePermissions(scheduleID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!permissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(permissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        taskScheduleService.finishScheduleManual(permissionResult.getResult(), manualScheduleStopInfo);

        return ResponseEntity.ok(new SimpleStatusResponse("success"));
    }
}
